@@ -2,9 +2,9 @@ use crate::csv::{CsvEncoding, NullValues};
 use crate::csv_core::utils::*;
 use crate::csv_core::{buffer::*, parser::*};
 use crate::mmap::MmapBytesReader;
+use crate::utils::resolve_homedir;
 use crate::PhysicalIoExpr;
 use crate::ScanAggregation;
-use csv::ByteRecordsIntoIter;
 use polars_arrow::array::*;
 use polars_core::utils::accumulate_dataframes_vertical;
 use polars_core::{prelude::*, POOL};
@@ -12,19 +12,21 @@ use rayon::prelude::*;
 use rayon::ThreadPoolBuilder;
 use std::borrow::Cow;
 use std::fmt;
+#[cfg(feature = "decompress")]
+use std::io::SeekFrom;
 use std::io::{Read, Seek};
 use std::path::PathBuf;
 use std::sync::atomic::Ordering;
 use std::sync::{atomic::AtomicUsize, Arc};
 
 /// CSV file reader
-pub struct SequentialReader<R: Read + MmapBytesReader> {
+pub struct CoreReader<R: Read + MmapBytesReader> {
     /// Explicit schema for the CSV file
     schema: SchemaRef,
     /// Optional projection for which columns to load (zero-based column indices)
     projection: Option<Vec<usize>>,
     /// File reader
-    record_iter: Option<ByteRecordsIntoIter<R>>,
+    reader: Option<R>,
     /// Current line number, used in error reporting
     line_number: usize,
     ignore_parser_errors: bool,
@@ -40,9 +42,12 @@ pub struct SequentialReader<R: Read + MmapBytesReader> {
     low_memory: bool,
     comment_char: Option<u8>,
     null_values: Option<Vec<String>>,
+    /// If schema was given by user or inferred
+    #[cfg(feature = "decompress")]
+    inferred_schema: bool,
 }
 
-impl<R> fmt::Debug for SequentialReader<R>
+impl<R> fmt::Debug for CoreReader<R>
 where
     R: Read + MmapBytesReader,
 {
@@ -103,7 +108,7 @@ impl RunningSize {
     }
 }
 
-impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
+impl<R: Read + Sync + Send + MmapBytesReader> CoreReader<R> {
     /// Returns the schema of the reader, useful for getting the schema without reading
     /// record batches
     pub fn schema(&self) -> SchemaRef {
@@ -230,29 +235,6 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
             }
         }
 
-        // assume 10 chars per str
-        // this is not updated in low memory mode
-        let init_str_bytes = chunk_size * 10;
-        let str_capacities: Vec<_> = str_columns
-            .iter()
-            .map(|_| RunningSize::new(init_str_bytes))
-            .collect();
-
-        // An empty file with a schema should return an empty DataFrame with that schema
-        if bytes.is_empty() {
-            let buffers = init_buffers(
-                &projection,
-                0,
-                &self.schema,
-                &str_capacities,
-                self.delimiter,
-            )?;
-            let df = DataFrame::new_no_checks(
-                buffers.into_iter().map(|buf| buf.into_series()).collect(),
-            );
-            return Ok(df);
-        }
-
         // split the file by the nearest new line characters such that every thread processes
         // approximately the same number of rows.
         let file_chunks =
@@ -277,6 +259,29 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
         // Structure:
         //      the inner vec has got buffers from all the columns.
         if predicate.is_some() {
+            // assume 10 chars per str
+            // this is not updated in low memory mode
+            let init_str_bytes = chunk_size * 10;
+            let str_capacities: Vec<_> = str_columns
+                .iter()
+                .map(|_| RunningSize::new(init_str_bytes))
+                .collect();
+
+            // An empty file with a schema should return an empty DataFrame with that schema
+            if bytes.is_empty() {
+                let buffers = init_buffers(
+                    &projection,
+                    0,
+                    &self.schema,
+                    &str_capacities,
+                    self.delimiter,
+                )?;
+                let df = DataFrame::new_no_checks(
+                    buffers.into_iter().map(|buf| buf.into_series()).collect(),
+                );
+                return Ok(df);
+            }
+
             let dfs = pool.install(|| {
                 file_chunks
                     .into_par_iter()
@@ -370,6 +375,21 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
             // let exponential growth solve the needed size. This leads to less memory overhead
             // in the later rechunk. Because we have large chunks they are easier reused for the
             // large final contiguous memory needed at the end.
+            let rows_per_thread = total_rows / n_threads;
+            let max_proxy = bytes.len() / n_threads / 2;
+            let capacity = if self.low_memory {
+                chunk_size
+            } else {
+                std::cmp::min(rows_per_thread, max_proxy)
+            };
+
+            // assume 10 chars per str
+            let init_str_bytes = capacity * 10;
+            let str_capacities: Vec<_> = str_columns
+                .iter()
+                .map(|_| RunningSize::new(init_str_bytes))
+                .collect();
+
             let dfs = pool.install(|| {
                 file_chunks
                     .into_par_iter()
@@ -382,7 +402,7 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
                         let mut read = bytes_offset_thread;
                         let mut buffers = init_buffers(
                             projection,
-                            chunk_size,
+                            capacity,
                             &schema,
                             &str_capacities,
                             self.delimiter,
@@ -406,7 +426,7 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
                                 self.encoding,
                                 // chunk size doesn't really matter anymore,
                                 // less calls if we increase the size
-                                chunk_size * 32,
+                                chunk_size * 320000,
                             )?;
                         }
                         Ok(DataFrame::new_no_checks(
@@ -419,6 +439,49 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
         }
     }
 
+    #[cfg(feature = "decompress")]
+    fn decompress_and_parse(
+        &mut self,
+        predicate: Option<Arc<dyn PhysicalIoExpr>>,
+        n_threads: usize,
+        bytes: &[u8],
+    ) -> Result<DataFrame> {
+        if let Some(bytes) = decompress(bytes) {
+            if self.inferred_schema {
+                self.schema = bytes_to_schema(
+                    &bytes,
+                    self.delimiter,
+                    self.has_header,
+                    self.skip_rows,
+                    self.comment_char,
+                )?;
+            }
+
+            self.parse_csv(n_threads, &bytes, predicate.as_ref())
+        } else {
+            self.parse_csv(n_threads, bytes, predicate.as_ref())
+        }
+    }
+
+    #[cfg(feature = "decompress")]
+    fn decompress<'a>(&mut self, bytes: &'a [u8]) -> Result<Cow<'a, [u8]>> {
+        if let Some(bytes) = decompress(bytes) {
+            if self.inferred_schema {
+                self.schema = bytes_to_schema(
+                    &bytes,
+                    self.delimiter,
+                    self.has_header,
+                    self.skip_rows,
+                    self.comment_char,
+                )?;
+            }
+
+            Ok(Cow::Owned(bytes))
+        } else {
+            Ok(Cow::Borrowed(bytes))
+        }
+    }
+
     /// Read the csv into a DataFrame. The predicate can come from a lazy physical plan.
     pub fn as_df(
         &mut self,
@@ -427,46 +490,76 @@ impl<R: Read + Sync + Send + MmapBytesReader> SequentialReader<R> {
     ) -> Result<DataFrame> {
         let n_threads = self.n_threads.unwrap_or_else(|| POOL.current_num_threads());
 
-        let mut df = match (&self.path, self.record_iter.is_some()) {
+        let mut df = match &self.path {
             // we have a path so we can mmap
-            (Some(p), _) => {
-                let file = std::fs::File::open(p)?;
+            Some(p) => {
+                let p = resolve_homedir(p);
+                let file = std::fs::File::open(&p)?;
                 let mmap = unsafe { memmap::Mmap::map(&file)? };
                 let bytes = mmap[..].as_ref();
+
+                #[cfg(feature = "decompress")]
+                {
+                    self.decompress_and_parse(predicate, n_threads, bytes)?
+                }
+
+                #[cfg(not(feature = "decompress"))]
                 self.parse_csv(n_threads, bytes, predicate.as_ref())?
             }
-            (None, true) => {
-                // get a hold of the reader + mmapreader
-                let mut r = std::mem::take(&mut self.record_iter)
-                    .unwrap()
-                    .into_reader()
-                    .into_inner();
+            None => {
+                let mut reader = self.reader.take().unwrap();
 
                 // we have a file so we can mmap
-                if let Some(file) = r.to_file() {
+                if let Some(file) = reader.to_file() {
                     let mmap = unsafe { memmap::Mmap::map(file)? };
                     let bytes = mmap[..].as_ref();
+
+                    #[cfg(feature = "decompress")]
+                    {
+                        self.decompress_and_parse(predicate, n_threads, bytes)?
+                    }
+
+                    #[cfg(not(feature = "decompress"))]
                     self.parse_csv(n_threads, bytes, predicate.as_ref())?
                 } else {
                     // we can get the bytes for free
-                    let bytes = if let Some(bytes) = r.to_bytes() {
+                    let bytes = if let Some(bytes) = reader.to_bytes() {
+                        #[cfg(feature = "decompress")]
+                        {
+                            self.decompress(bytes)?
+                        }
+
+                        #[cfg(not(feature = "decompress"))]
                         Cow::Borrowed(bytes)
                     } else {
                         // we have to read to an owned buffer to get the bytes.
                         let mut bytes = Vec::with_capacity(1024 * 128);
-                        r.read_to_end(&mut bytes)?;
+                        reader.read_to_end(&mut bytes)?;
                         if !bytes.is_empty()
                             && (bytes[bytes.len() - 1] != b'\n' || bytes[bytes.len() - 1] != b'\r')
                         {
                             bytes.push(b'\n')
                         }
+
+                        #[cfg(feature = "decompress")]
+                        {
+                            match self.decompress(&bytes)? {
+                                // always return the owned version
+                                Cow::Borrowed(_) => Cow::Owned(bytes),
+                                Cow::Owned(bytes) => Cow::Owned(bytes),
+                            }
+                        }
+
+                        #[cfg(not(feature = "decompress"))]
                         Cow::Owned(bytes)
                     };
 
-                    self.parse_csv(n_threads, &bytes, predicate.as_ref())?
+                    let out = self.parse_csv(n_threads, &bytes, predicate.as_ref())?;
+                    // restore reader for a potential second run
+                    self.reader = Some(reader);
+                    out
                 }
             }
-            _ => return Err(PolarsError::Other("file or reader must be set".into())),
         };
 
         if let Some(aggregate) = aggregate {
@@ -509,22 +602,56 @@ pub fn build_csv_reader<R: Read + Seek + Sync + Send + MmapBytesReader>(
     low_memory: bool,
     comment_char: Option<u8>,
     null_values: Option<NullValues>,
-) -> Result<SequentialReader<R>> {
+) -> Result<CoreReader<R>> {
     // check if schema should be inferred
     let delimiter = delimiter.unwrap_or(b',');
+
+    #[cfg(feature = "decompress")]
+    let mut inferred_schema = false;
+
     let schema = match schema {
         Some(schema) => schema,
         None => {
-            let (inferred_schema, _) = infer_file_schema(
-                &mut reader,
-                delimiter,
-                max_records,
-                has_header,
-                schema_overwrite,
-                skip_rows,
-                comment_char,
-            )?;
-            Arc::new(inferred_schema)
+            #[cfg(feature = "decompress")]
+            {
+                // We keep track of the inferred schema bool
+                // In case the file is compressed this schema inference is wrong and has to be done
+                // again after decompression.
+                inferred_schema = true;
+
+                const N: usize = 5;
+                let mut bytes = [0u8; N];
+                reader.read_exact(&mut bytes)?;
+                // restore position
+                reader.seek(SeekFrom::Current(-(N as i64)))?;
+                if decompress(&bytes).is_some() {
+                    Arc::new(Schema::default())
+                } else {
+                    let (inferred_schema, _) = infer_file_schema(
+                        &mut reader,
+                        delimiter,
+                        max_records,
+                        has_header,
+                        schema_overwrite,
+                        skip_rows,
+                        comment_char,
+                    )?;
+                    Arc::new(inferred_schema)
+                }
+            }
+            #[cfg(not(feature = "decompress"))]
+            {
+                let (inferred_schema, _) = infer_file_schema(
+                    &mut reader,
+                    delimiter,
+                    max_records,
+                    has_header,
+                    schema_overwrite,
+                    skip_rows,
+                    comment_char,
+                )?;
+                Arc::new(inferred_schema)
+            }
         }
     };
 
@@ -539,13 +666,10 @@ pub fn build_csv_reader<R: Read + Seek + Sync + Send + MmapBytesReader>(
         projection = Some(prj);
     }
 
-    let csv_reader = init_csv_reader(reader, has_header, delimiter, comment_char);
-    let record_iter = Some(csv_reader.into_byte_records());
-
-    Ok(SequentialReader {
+    Ok(CoreReader {
         schema,
         projection,
-        record_iter,
+        reader: Some(reader),
         line_number: if has_header { 1 } else { 0 },
         ignore_parser_errors,
         skip_rows,
@@ -560,5 +684,7 @@ pub fn build_csv_reader<R: Read + Seek + Sync + Send + MmapBytesReader>(
         low_memory,
         comment_char,
         null_values,
+        #[cfg(feature = "decompress")]
+        inferred_schema,
     })
 }
